import numpy as np
from pygt3x.bit_pack_acceleration import BitPackAcceleration


class Activity1Payload:
    """
    Class for Activity 1 Payload
    """

    def __init__(self, payload_bytes, timestamp):
        self.AccelerationSamples = BitPackAcceleration.unpack_activity(
            payload_bytes, timestamp, True
        )


class Activity2Payload:
    """Class for Activity 2 Payload."""

    @staticmethod
    def unpack_activity2(payload_bytes, timestamp):
        if (len(payload_bytes) % 6) != 0:
            payload_bytes = payload_bytes[: -(len(payload_bytes) % 6) + 1]
        data = np.frombuffer(payload_bytes, dtype=np.int16).reshape((-1, 3))
        data = np.concatenate(
            (np.array(timestamp).repeat(len(data)).reshape(-1, 1), data), axis=1
        )
        return data.reshape((-1, 4))

    def __init__(self, payload_bytes, timestamp):
        self.AccelerationSamples = self.unpack_activity2(payload_bytes, timestamp)


class Activity3Payload:
    """
    Class for Activity 3 Payload
    """

    def __init__(self, payload_bytes, timestamp):
        self.AccelerationSamples = BitPackAcceleration.unpack_activity(
            payload_bytes, timestamp, False
        )



import io
import json
from zipfile import ZipFile

import numpy as np
import pandas as pd

from pygt3x.activity_payload import Activity1Payload
from pygt3x.activity_payload import Activity2Payload
from pygt3x.activity_payload import Activity3Payload
from pygt3x.componenets import Info, RawEvent, Header
from pygt3x.events import Types


class FileReader:
    """
    Class for Gt3x file reader

    Reads GT3X/AGDC files
    """

    def __init__(self, file_name):
        self.file_name = file_name

    def __enter__(self):
        self.zipfile = ZipFile(self.file_name)
        self.logfile = self.zipfile.open("log.bin", "r")
        self.logreader = LogReader(self.logfile)
        return self

    def __exit__(self, typ, value, traceback):
        self.logfile.__exit__(typ, value, traceback)
        self.zipfile.__exit__(typ, value, traceback)

    def read_info(self):
        """
        Parses info.txt and returns dictionary with key/value pairs
        """
        output = dict()
        with io.TextIOWrapper(
            self.zipfile.open("info.txt", "r"), encoding="utf-8-sig"
        ) as infoFile:
            for line in infoFile.readlines():
                values = line.split(":")
                if len(values) == 2:
                    output[values[0].strip()] = values[1].strip()

        return Info(output)

    def read_calibration(self):
        if "calibration.json" not in self.zipfile.namelist():
            return None
        with self.zipfile.open("calibration.json") as calibrationFile:
            calibration = json.load(calibrationFile)
            return calibration

    def read_events(self, num_rows=None):
        if num_rows is None:
            raw_event = self.logreader.read_event()
            while raw_event is not None:
                yield raw_event
                raw_event = self.logreader.read_event()
        else:
            for _ in range(0, num_rows):
                raw_event = self.logreader.read_event()
                yield raw_event

    def get_acceleration(self, num_rows=None):
        for evt in self.read_events(num_rows):
            if Types(evt.header.eventType) == Types.Activity3:
                payload = Activity3Payload(evt.payload, evt.header.timestamp)
            elif Types(evt.header.eventType) == Types.Activity2:
                payload = Activity2Payload(evt.payload, evt.header.timestamp)
            elif Types(evt.header.eventType) == Types.Activity:
                payload = Activity1Payload(evt.payload, evt.header.timestamp)
            else:
                continue

            yield payload.AccelerationSamples

    def to_pandas(self):
        """
        Returns acceleration data as pandas data frame
        """
        col_names = ["Timestamp", "X", "Y", "Z"]
        data = np.concatenate(list(self.get_acceleration()))
        df = pd.DataFrame(data, columns=col_names)
        df.index = df["Timestamp"]
        del df["Timestamp"]
        return df


class LogReader:
    """
    Class to handle reading GT3X/AGDC log events
    """

    def __init__(self, source):
        """
        Constructor for Gt3xLogReader

        Parameters:
        source: IO stream for log.bin file data

        """
        self.source = source

    def read_event(self):
        header_bytes = self.source.read(8)
        if len(header_bytes) != 8:
            return None
        header = Header(header_bytes)
        payload_bytes = self.source.read(header.payload_size)
        if len(payload_bytes) != header.payload_size:
            return None
        checksum = self.source.read(1)
        if not checksum:
            return None
        raw_event = RawEvent(header, payload_bytes, checksum)
        return raw_event



import numpy as np


class CalibrationV2Service:
    def __init__(self, calibration: dict, sample_rate: int):
        self.offset_vector = np.array([[0, 0, 0]])
        self.sensitivity_matrix = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        self.set_calibration(calibration, sample_rate)

    def set_calibration(self, calibration: dict, sample_rate: int):
        offset_x = float(calibration[f"offsetX_{sample_rate}"])
        offset_y = float(calibration[f"offsetY_{sample_rate}"])
        offset_z = float(calibration[f"offsetZ_{sample_rate}"])

        sensitivity_xx = float(calibration[f"sensitivityXX_{sample_rate}"])
        sensitivity_yy = float(calibration[f"sensitivityYY_{sample_rate}"])
        sensitivity_zz = float(calibration[f"sensitivityZZ_{sample_rate}"])

        sensitivity_xy = float(calibration[f"sensitivityXY_{sample_rate}"])
        sensitivity_xz = float(calibration[f"sensitivityXZ_{sample_rate}"])
        sensitivity_yz = float(calibration[f"sensitivityYZ_{sample_rate}"])

        s11 = (sensitivity_xx * 0.01) ** -1.0
        s12 = ((sensitivity_xy * 0.01 + 250) ** -1.0) - 0.004
        s13 = ((sensitivity_xz * 0.01 + 250) ** -1.0) - 0.004

        s21 = ((sensitivity_xy * 0.01 + 250) ** -1.0) - 0.004
        s22 = (sensitivity_yy * 0.01) ** -1.0
        s23 = ((sensitivity_yz * 0.01 + 250) ** -1.0) - 0.004

        s31 = ((sensitivity_xz * 0.01 + 250) ** -1.0) - 0.004
        s32 = ((sensitivity_yz * 0.01 + 250) ** -1.0) - 0.004
        s33 = (sensitivity_zz * 0.01) ** -1.0

        self.offset_vector = np.array([[offset_x, offset_y, offset_z]])
        self.sensitivity_matrix = np.array(
            [[s11, s21, s31], [s12, s22, s32], [s13, s23, s33]]
        )

    def calibrate_samples(self, sample):
        return np.concatenate(
            (
                sample[:, :-3],
                np.matmul(
                    self.sensitivity_matrix,
                    (sample[:, -3:] - self.offset_vector).transpose(),
                ).transpose(),
            ),
            axis=1,
        )



"""Class for unpacking activity."""
import numpy as np


class BitPackAcceleration:
    @staticmethod
    def unpack_activity(source, timestamp: int, flip_xy: bool):
        """
        Unpacks activity stored as sets of 3, 12-bit integers

        Parameters:
        source (byte array): Activity payload bytes array
        flipXY (bool): If true sample is extracted in Y,X,Z order.
         Otherwise X,Y,Z order.

        Returns:
        Generator which produces acceleration samples as Int16 values

        """
        data = np.frombuffer(source, dtype=np.uint8)
        fst_uint8, mid_uint8, lst_uint8 = (
            np.reshape(data, (data.shape[0] // 3, 3)).astype(np.int16).T
        )
        fst_uint12 = (fst_uint8 << 4) + (mid_uint8 >> 4)
        snd_uint12 = ((mid_uint8 % 16) << 8) + lst_uint8
        concat = np.concatenate((fst_uint12[:, None], snd_uint12[:, None]), axis=1)
        data = concat.reshape((-1, 3))
        data[data > 2047] = data[data > 2047] - 4095
        data = np.concatenate(
            (np.array(timestamp).repeat(len(data)).reshape(-1, 1), data), axis=1
        )
        if flip_xy:
            data = data[:, [1, 0, 2]]
        return data



from typing import Union

import numpy as np
import pandas as pd

from pygt3x.activity_payload import Activity1Payload
from pygt3x.activity_payload import Activity2Payload
from pygt3x.activity_payload import Activity3Payload
from pygt3x.callibration_v2_service import CalibrationV2Service
from pygt3x.componenets import AccelerationSample
from pygt3x.componenets import RawEvent
from pygt3x.events import Types
from pygt3x.reader import FileReader


class CalibratedReader:
    """Calibrated event reader.

    Will calibrate activity events as they are read.
    """

    def __init__(self, source: FileReader):
        self.source = source

    def read_info(self):
        """
        Returns info dictionary from source reader
        """
        return self.source.read_info()

    def read_calibration(self):
        return self.source.read_calibration()

    def calibrate_acceleration(self, raw_event: RawEvent):
        """
        Calibrates acceleration samples.

        Parameters:
            raw_event (RawEvent): Activity event to calibrate.

        """
        payload: Union[Activity1Payload, Activity2Payload, Activity3Payload]
        if Types(raw_event.header.eventType) == Types.Activity3:
            payload = Activity3Payload(raw_event.payload, raw_event.header.timestamp)
        elif Types(raw_event.header.eventType) == Types.Activity:
            payload = Activity1Payload(raw_event.payload, raw_event.header.timestamp)
        elif Types(raw_event.header.eventType) == Types.Activity2:
            payload = Activity2Payload(raw_event.payload, raw_event.header.timestamp)
        else:
            raise ValueError("Cannot calibrate non-activity event type")

        calibration = self.source.read_calibration()
        info = self.read_info()

        if calibration is None or calibration["isCalibrated"]:
            # Data is already calibrated, so just return unscaled values
            accel_scale = info.get_acceleration_scale()
            raw_event.calibrated_acceleration = np.concatenate(
                (
                    payload.AccelerationSamples[:, :-3],
                    payload.AccelerationSamples[:, -3:] / accel_scale,
                ),
                axis=1,
            )
        elif calibration["calibrationMethod"] == 2:
            # Use calibration method 2 to calibrate activity
            sample_rate = info.get_sample_rate()
            raw_event.calibrated_acceleration = self.calibrate_v2(
                payload.AccelerationSamples, calibration, sample_rate
            )
        else:
            raise NotImplementedError(
                f"Unknown calibration method: " f"{calibration['calibrationMethod']}"
            )

    @staticmethod
    def calibrate_v2(samples, calibration: dict, sample_rate: int):
        calibration_service = CalibrationV2Service(calibration, sample_rate)
        return calibration_service.calibrate_samples(samples)

    def read_events(self, num_rows: int = None):
        """
        Read events from source and calibrates activity.

        Parameters:
            num_rows (int): Optionally limits number of rows to return.

        """
        for raw_event in self.source.read_events(num_rows):
            if not Types(raw_event.header.eventType) in [
                Types.Activity,
                Types.Activity2,
                Types.Activity3,
            ]:
                continue
            self.calibrate_acceleration(raw_event)
            yield raw_event

    def get_samples(self):
        for raw_event in self.read_events():
            yield raw_event.calibrated_acceleration

    def to_pandas(self):
        """
        Returns acceleration data as pandas data frame
        """
        col_names = ["Timestamp", "X", "Y", "Z"]
        data = np.concatenate(list(self.get_samples()))
        df = pd.DataFrame(data, columns=col_names)
        df.index = df["Timestamp"]
        del df["Timestamp"]
        return df



from enum import Enum, unique


@unique
class Types(Enum):
    Activity = 0
    AntPlus = 1
    Battery = 2
    Event = 3
    HeartRateBpm = 4
    Lux = 5
    MetaData = 6
    Tag = 7
    Temperature = 8
    Epoch = 9
    DailySummary = 10
    HeartRateAnt = 11
    Epoch2 = 12
    Capsense = 13
    HeartRateBle = 14
    Epoch3 = 15
    Epoch4 = 16
    Imu9dof = 17
    Magnetometer = 18
    FifoError = 19
    FifoDump = 20
    Params = 21
    ImuAccelerometer = 22
    Debug = 23
    SensorSchema = 24
    SensorData = 25
    Activity2 = 26
    Activity3 = 27
    LongPress = 28
    ButtonEvent = 29



"""GT3x header structure."""
import struct
from typing import List


class AccelerationSample:
    """
    Acceleration Sample Model

    """

    def __init__(self, timestamp: int, x: float, y: float, z: float):
        self.timestamp = timestamp
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return f"Timestamp: {self.timestamp}, X: {self.x}," f" Y: {self.y}, Z: {self.z}"

    def __iter__(self):
        return iter([self.timestamp, self.x, self.y, self.z])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.timestamp == other.timestamp
                and self.x == other.x
                and self.y == other.y
                and self.z == other.z
            )
        else:
            return False

    def __sub__(self, other):
        return AccelerationSample(
            self.timestamp, self.x - other.x, self.y - other.y, self.z - other.z
        )

    def __abs__(self):
        return AccelerationSample(self.timestamp, abs(self.x), abs(self.y), abs(self.z))


class Header:
    """
    Class for Gt3xHeader

    Attributes:
        separator (byte): Log separator value
        timestamp (long): Unix epoch timestamp in seconds
        eventType (byte): GT3X event type
        payload_size (int): Event payload size in bytes

    """

    def __init__(self, header_bytes: bytes):
        (separator, eventType, timestamp, payload_size) = struct.unpack(
            "<BBLH", header_bytes
        )
        self.separator = separator
        self.timestamp = timestamp
        self.eventType = eventType
        self.payload_size = payload_size


class RawEvent:
    """
    Class for Gt3xRawEvent

    Parameters:
        header: Gt3xHeader
        payload (byte array): Log event payload as byte array
        checksum (byte): Log event checksum

    Attributes:
        calibrated_acceleration: List[AccelerationSample]
    """

    calibrated_acceleration: List[AccelerationSample] = []

    def __init__(self, header: Header, payload, checksum):
        self.header = header
        self.payload = payload
        self.checksum = checksum


class Info(dict):
    def get_sample_rate(self):
        return int(self["Sample Rate"])

    def get_acceleration_scale(self):
        return float(self["Acceleration Scale"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pygt3x']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.2,<2.0.0', 'pandas>=1.0.5,<2.0.0']

setup_kwargs = {
    'name': 'pygt3x',
    'version': '0.1.3',
    'description': 'Python module for reading GT3X/AGDC file format data',
    'long_description': '# pygt3x\n![Tests](https://github.com/actigraph/pygt3x/actions/workflows/tests.yml/badge.svg)\n\nPython module for reading GT3X/AGDC file format data generated by ActiGraph devices.\n\n## Example Usage\n\n```python\nfrom pygt3x.reader import FileReader\nfrom pygt3x.calibrated_reader import CalibratedReader\n\n# Read raw data and calibrate\n# Dump to pandas data frame\nwith FileReader("WRIST_rawLSB_032Hz.agdc") as reader:\n    calibrated_reader =  CalibratedReader(reader)\n    df = calibrated_reader.to_pandas()\n    print(df.head(5))\n```\n',
    'author': 'Mark Fogle',
    'author_email': 'mark.fogle@theactigraph.com',
    'maintainer': 'Ali Neishabouri',
    'maintainer_email': 'ali.neishabouri@theactigraph.com',
    'url': 'https://github.com/actigraph/pygt3x',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)
